#!perl

use strict;
use warnings FATAL => 'all';

use lib "lib";

use Archive::Tar;
use Carp;
use Carp::Assert;
use Cwd;
use Data::Dumper;
use English qw{-no_match_vars};
use Test::More;
use Test::BDD::Cucumber::StepFile;
use Method::Signatures::Simple;

use List::Util qw{first};
use Path::Class;
use File::Temp qw{tempdir};

use Tails::IUK;
use Tails::IUK::Read;
use Tails::IUK::Utils qw{get_temp_dir};

my $bindir = file(__FILE__)->dir->parent->parent->parent->subdir('bin')->absolute;
use Env qw{@PATH};
unshift @PATH, $bindir;

Given qr{^a usable temporary directory$}, func($c) {
    my $dirname = get_temp_dir();
    $c->{stash}->{scenario}->{tempdir} = $dirname;
    ok(defined($dirname));
};

Given qr{^the "([^"]+)" empty SquashFS diff$}, func($c) {
    my $basename = $c->matches->[0];
    $c->{stash}->{scenario}->{squashfs_diff} = file($c->{stash}->{scenario}->{tempdir}, $basename);
    ok($c->{stash}->{scenario}->{squashfs_diff}->touch);
};

func inject_new_bootloader_bits_into($dir) {
    for (qw{EFI/BOOT/bootx64.efi utils/linux/syslinux}) {
        my $injected_file = file($dir, $_);
        $injected_file->parent->mkpath;
        $injected_file->touch;
        ok(-e $injected_file);
    }
}

func geniso($srcdir, $outfile) {
    dir($srcdir, 'isolinux')->mkpath;
    assert(-d dir($srcdir, 'isolinux'));
    my $isolinux_cfg_fh = file($srcdir, 'isolinux', 'isolinux.cfg')->openw;
    print $isolinux_cfg_fh "bla\n   bli/isolinux/blu\n\n\n bla/isolinux/";
    close $isolinux_cfg_fh;
    assert(-e file($srcdir, 'isolinux', 'isolinux.cfg'));

    dir($srcdir, 'live')->mkpath;
    if (! -e file($srcdir, 'live', 'filesystem.squashfs')) {
        my $squashfs_tempdir = tempdir(CLEANUP => 1);
        # an empty SquashFS is invalid
        file($squashfs_tempdir, '.placeholder')->touch;
        `mksquashfs $squashfs_tempdir $srcdir/live/filesystem.squashfs -no-progress 2>/dev/null`;
        ${^CHILD_ERROR_NATIVE} == 0 or croak("Failed to build SquashFS");
    }
    `genisoimage --quiet -J -l -cache-inodes -allow-multidot -o "$outfile" "$srcdir" 2>/dev/null`;
    ${^CHILD_ERROR_NATIVE} == 0
}

Given qr{^(an old|a new) ISO image(?: that does not contain file "([^"]+)")?$}, func($c) {
    my $generation = $c->matches->[0] eq 'an old' ? 'old' : 'new';
    my $file = $c->matches->[1];
    my $basename = $generation eq 'old' ? 'old.iso' : 'new.iso';
    my $filename = dir($c->{stash}->{scenario}->{tempdir}, $basename);
    my $iso_tempdir = tempdir(CLEANUP => 1);
    if (defined $file) {
        assert(length $file);
        ok(! -e file($iso_tempdir, $file));
    }
    inject_new_bootloader_bits_into($iso_tempdir) if $generation eq 'new';
    ok(geniso($iso_tempdir, $filename));
};

Given qr{^(an old|a new) ISO image that contains file "([^"]+)"$}, func($c) {
    my $generation = $c->matches->[0];
    my $file = $c->matches->[1];
    my $basename = $generation eq 'an old' ? 'old.iso' : 'new.iso';
    my $filename = dir($c->{stash}->{scenario}->{tempdir}, $basename);
    my $iso_tempdir = tempdir(CLEANUP => 1);
    file($iso_tempdir, $file)->dir->mkpath();
    file($iso_tempdir, $file)->touch;
    ok(-e file($iso_tempdir, $file));
    inject_new_bootloader_bits_into($iso_tempdir) if $generation eq 'new';
    ok(geniso($iso_tempdir, $filename));
};

Given qr{^(?:two identical ISO images|two ISO images that contain the same set of kernels|two ISO images that contain the same bootloader configuration)$}, func($c) {
    for my $generation (qw{old new}) {
        my $basename = $generation.'.iso';
        my $filename = dir($c->{stash}->{scenario}->{tempdir}, $basename);
        my $iso_tempdir = tempdir(CLEANUP => 1);
        dir($iso_tempdir, 'live')->mkpath;
        inject_new_bootloader_bits_into($iso_tempdir);
        geniso($iso_tempdir, $filename) or croak "Failed to generate ISO image.";
    }
    ok(
        -f dir($c->{stash}->{scenario}->{tempdir}, 'old.iso')
     && -f dir($c->{stash}->{scenario}->{tempdir}, 'new.iso')
    );
};

Given qr{^two ISO images when the kernel was upgraded$}, func($c) {
    for my $generation (qw{old new}) {
        my $basename = $generation.'.iso';
        my $filename = dir($c->{stash}->{scenario}->{tempdir}, $basename);
        my $iso_tempdir = tempdir(CLEANUP => 1);
        dir($iso_tempdir, 'live')->mkpath;
        for (qw{vmlinuz initrd.img}) {
            my $fh = file($iso_tempdir, 'live', $_)->openw;
            print $fh $generation; close $fh;
        }
        inject_new_bootloader_bits_into($iso_tempdir) if $generation eq 'new';
        geniso($iso_tempdir, $filename) or croak "Failed to generate ISO image.";
    }

    ok(
        -f dir($c->{stash}->{scenario}->{tempdir}, 'old.iso')
     && -f dir($c->{stash}->{scenario}->{tempdir}, 'new.iso')
    );
};

Given qr{^two ISO images when a new kernel was added$}, func($c) {
    for my $generation (qw{old new}) {
        my $basename = $generation.'.iso';
        my $filename = dir($c->{stash}->{scenario}->{tempdir}, $basename);
        my $iso_tempdir = tempdir(CLEANUP => 1);
        dir($iso_tempdir, 'live')->mkpath;
        for (qw{vmlinuz initrd.img}) {
            my $fh = file($iso_tempdir, 'live', $_)->openw;
            print $fh "same content"; close $fh;
        }
        if ($generation eq 'new') {
            for (qw{vmlinuz2 initrd2.img}) {
                my $fh = file($iso_tempdir, 'live', $_)->openw;
                print $fh $generation; close $fh;
            }
            inject_new_bootloader_bits_into($iso_tempdir);
        }
        geniso($iso_tempdir, $filename) or croak "Failed to generate ISO image.";
    }

    ok(
        -f dir($c->{stash}->{scenario}->{tempdir}, 'old.iso')
     && -f dir($c->{stash}->{scenario}->{tempdir}, 'new.iso')
    );
};

Given qr{^(an old|a new) ISO image whose filesystem.squashfs( does not|) contains? file "([^"]+)"(?:| modified at ([0-9]+))$}, func($c) {
    my $generation = $c->matches->[0] eq 'an old' ? 'old' : 'new';
    my $contains = $c->matches->[1] eq "" ? 1 : 0;
    my $file     = $c->matches->[2];
    my $mtime    = $c->matches->[3];

    my $iso_basename = $generation eq 'old' ? 'old.iso' : 'new.iso';
    my $iso_filename = file($c->{stash}->{scenario}->{tempdir}, $iso_basename);
    my $iso_tempdir = tempdir(CLEANUP => 1);
    my $squashfs_tempdir = tempdir(CLEANUP => 1);
    # an empty SquashFS is invalid
    file($squashfs_tempdir, '.placeholder')->touch;
    if ($contains) {
        file($squashfs_tempdir, $file)->dir->mkpath();
        file($squashfs_tempdir, $file)->touch;
        utime($mtime, $mtime, file($squashfs_tempdir, $file)) if defined($mtime);
    }
    dir($iso_tempdir, 'live')->mkpath();
    `mksquashfs $squashfs_tempdir $iso_tempdir/live/filesystem.squashfs -no-progress 2>/dev/null`;
    ${^CHILD_ERROR_NATIVE} == 0 or croak("Failed to build SquashFS");
    inject_new_bootloader_bits_into($iso_tempdir) if $generation eq 'new';
    ok(geniso($iso_tempdir, $iso_filename));
};

Given qr{^two ISO images that do not contain the same bootloader configuration$}, func($c) {
    for my $generation (qw{old new}) {
        my $basename = $generation.'.iso';
        my $filename = dir($c->{stash}->{scenario}->{tempdir}, $basename);
        my $iso_tempdir = tempdir(CLEANUP => 1);
        dir($iso_tempdir, 'isolinux')->mkpath;
        for (qw{live.cfg}) {
            my $fh = file($iso_tempdir, 'isolinux', $_)->openw;
            print $fh $generation; close $fh;
        }
        inject_new_bootloader_bits_into($iso_tempdir) if $generation eq 'new';
        geniso($iso_tempdir, $filename) or croak "Failed to generate ISO image.";
    }

    ok(
        -f dir($c->{stash}->{scenario}->{tempdir}, 'old.iso')
     && -f dir($c->{stash}->{scenario}->{tempdir}, 'new.iso')
    );
};

Given qr{^the "([^"]+)" (compressed |)archive that contains file "([^"]+)"$}, func($c) {
    my $outfile = file($c->{stash}->{scenario}->{tempdir}, $c->matches->[0]);
    my $compressed = $c->matches->[1] ? 1 : 0;
    my $infile  = $c->matches->[2];

    my $tempdir = get_temp_dir();
    my $fh = file($tempdir, $infile)->openw;
    print $fh "something";
    close $fh;

    my $tarball = Archive::Tar->new();
    chdir $tempdir;
    $tarball->add_files($infile);
    my @args = ($outfile);
    push @args, COMPRESS_BZIP if $compressed;
    $tarball->write(@args);

    $c->{stash}->{scenario}->{tarballs} = [] unless exists $c->{stash}->{scenario}->{tarballs};
    push @{$c->{stash}->{scenario}->{tarballs}}, $outfile->stringify;
};

When qr{^I create an IUK$}, func($c) {
    my %args;
    $c->{stash}->{scenario}->{squashfs_diff_name} =
        'Tails_i386_0.11.1_to_0.11.2.squashfs';

    my $iuk_filename = file($c->{stash}->{scenario}->{tempdir}, 'test.iuk');

    my $cmdline =
        file($bindir, "tails-create-iuk") .
        ' --old-iso "' .
        file($c->{stash}->{scenario}->{tempdir}, 'old.iso') . '" ' .
        ' --new-iso "' .
        file($c->{stash}->{scenario}->{tempdir}, 'new.iso') . '"' .
        ' --squashfs-diff-name "'.$c->{stash}->{scenario}->{squashfs_diff_name}.'"' .
        ' --outfile "' . $iuk_filename . '"'
    ;
    if (exists $c->{stash}->{scenario}->{tarballs}) {
        foreach (@{$c->{stash}->{scenario}->{tarballs}}) {
            $cmdline .= ' --tarballs "' . $_ . '"';
        }
    }
    if (exists $c->{stash}->{scenario}->{squashfs_diff}) {
        $cmdline .=
            ' --squashfs-diff "' . $c->{stash}->{scenario}->{squashfs_diff} . '"';
    }

    $c->{stash}->{scenario}->{create_output} = `umask 077 && $cmdline 2>&1`;
    $c->{stash}->{scenario}->{create_exit_code} = ${^CHILD_ERROR_NATIVE};

    $c->{stash}->{scenario}->{create_exit_code} == 0
        or warn $c->{stash}->{scenario}->{create_output};

    my $iuk_in = Tails::IUK::Read->new_from_file($iuk_filename);
    $c->{stash}->{scenario}->{iuk_in} = $iuk_in;
    ok(defined($iuk_in));
};

Then qr{^the saved IUK should contain a "([^"]*)" file$}, func($c) {
    my $file = $c->matches->[0];
    ok($c->{stash}->{scenario}->{iuk_in}->contains_file($file));
};

func archive_contains_only_files_owned_by ($archive_filename, $owner, $group) {
    map { like(
        $_,
        qr{
             \A            # at the beginning of the string
             [-a-z]+       # permissions
             [[:space:]]+
             $owner        # owner
             /
             $group        # group
             [[:space:]]+
         }xms,
        "line looks like a file description with owner $owner and group $group"
    ) } split(/\n/, `tar tv -f '$archive_filename'`);
}

Then qr{^all files in the saved IUK should belong to owner 0 and group 0$}, func($c) {
    ok(archive_contains_only_files_owned_by(
        $c->{stash}->{scenario}->{iuk_in}->file,
        0,
        0,
    ));
};

Then qr{all files in the "([^"]+)" archive should belong to owner 0 and group 0$}, func($c) {
    my $archive_name = $c->matches->[0];
    my $orig_cwd = getcwd;
    chdir tempdir();
    system(qw{tar xf}, $c->{stash}->{scenario}->{iuk_in}->file, $archive_name);
    ${^CHILD_ERROR_NATIVE} == 0 or croak("Failed to extract IUK");
    ok(archive_contains_only_files_owned_by(
        $archive_name,
        0,
        0,
    ));
    chdir $orig_cwd;
};

Then qr{^the "([^"]+)" file in the saved IUK should contain "([^"]*)"$}, func($c) {
    my $file = $c->matches->[0];
    my $expected_content = $c->matches->[1];
    is($c->{stash}->{scenario}->{iuk_in}->get_content($file), $expected_content);
};

func archive_in_iuk_contains($iuk_in, $archive_name, $expected_file) {
    my $orig_cwd = getcwd;
    chdir tempdir();
    system(qw{tar xf}, $iuk_in->file);
    ${^CHILD_ERROR_NATIVE} == 0 or croak("Failed to extract IUK");
    my $archive = Archive::Tar->new($archive_name)
        or croak "Could not create Archive::Tar object ($archive_name)";
    chdir $orig_cwd;
    # contains_file is too fragile
    1 == grep { $_ eq $expected_file } $archive->list_files();
}

func _file_content_in_archive_in_iuk_like(
    $iuk_in, $archive_name, $filename, $regexp, $should_match
) {
    assert(defined $iuk_in && defined $archive_name && defined $filename);
    assert(defined $regexp && defined $should_match);

    my $orig_cwd = getcwd;
    chdir tempdir();
    system(qw{tar xf}, $iuk_in->file);
    ${^CHILD_ERROR_NATIVE} == 0 or croak("Failed to extract IUK");
    my $archive = Archive::Tar->new($archive_name);
    chdir $orig_cwd;

    # contains_file is too fragile
    unless (1 == grep { $_ eq $filename } $archive->list_files()) {
        warn "$archive_name does not contain $filename, so we can't check its content.";
        return;
    }

    my $content = $archive->get_content($filename);

    if ($should_match) {
        return $content =~ m{$regexp}xms;
    }
    else {
        return $content !~ m{$regexp}xms;
    }
}

func file_content_in_archive_in_iuk_like($iuk_in, $archive_name, $filename, $regexp) {
    assert(defined $iuk_in && defined $archive_name && defined $filename && defined $regexp);
    _file_content_in_archive_in_iuk_like(@_, 1);
}

func file_content_in_archive_in_iuk_unlike($iuk_in, $archive_name, $filename, $regexp) {
    assert(defined $iuk_in && defined $archive_name && defined $filename && defined $regexp);
    _file_content_in_archive_in_iuk_like(@_, 0);
}

func squashfs_in_iuk_contains($iuk_in, $squashfs_name, $expected_file, $expected_mtime) {
    $squashfs_name = file('live', $squashfs_name);
    my $orig_cwd = getcwd;
    my $tempdir = tempdir(CLEANUP => 1);
    chdir $tempdir;
    system(qw{tar xf}, $iuk_in->file);
    ${^CHILD_ERROR_NATIVE} == 0 or croak("Failed to extract IUK");
    die "system.tar not found in IUK" unless -e 'system.tar';
    my $system_tar = Archive::Tar->new('system.tar');
    $system_tar->extract();
    die "SquashFS '$squashfs_name' not found in system.tar" unless -e $squashfs_name;
    system(
        "unsquashfs", '-no-progress',
        file($tempdir, $squashfs_name),
        $expected_file
    );
    my $exists = ${^CHILD_ERROR_NATIVE} == 0 ? 1 : 0;
    chdir $orig_cwd;

    return unless $exists;

    if (defined $expected_mtime) {
        return $expected_mtime == file($tempdir, 'squashfs-root', $expected_file)->stat->mtime
    }
    return 1;
}

func squashfs_in_iuk_deletes($iuk_in, $squashfs_name, $deleted_file) {
    $squashfs_name = file('live', $squashfs_name);
    my $orig_cwd = getcwd;
    my $tempdir = tempdir(CLEANUP => 1);
    chdir $tempdir;
    system(qw{tar xf}, $iuk_in->file);
    ${^CHILD_ERROR_NATIVE} == 0 or croak("Failed to extract IUK");
    die "system.tar not found in IUK" unless -e 'system.tar';
    my $system_tar = Archive::Tar->new('system.tar');
    $system_tar->extract();
    die "SquashFS '$squashfs_name' not found in system.tar" unless -e $squashfs_name;

    my $old_dir = tempdir(CLEANUP => 1);
    file($old_dir, $deleted_file)->touch;

    my $new_dir = tempdir(CLEANUP => 1);
    system("unsquashfs", "-force", "-dest", $new_dir, $squashfs_name);
    ${^CHILD_ERROR_NATIVE} == 0 or croak("Failed to extract SquashFS.");

    my $union_mountpoint = tempdir(CLEANUP => 1);
    system(
        qw{sudo -n sudo mount -t aufs},
        "-o", sprintf("noatime,noxino,br=%s=rw:%s=rr+wh", $new_dir, $old_dir),
        $new_dir,
        $union_mountpoint
    );
    ${^CHILD_ERROR_NATIVE} == 0 or croak("Failed to union-mount.");

    my $exists = -e file($union_mountpoint, $deleted_file);
    system(qw{sudo umount}, "$union_mountpoint");
    return ! $exists;
}

Then qr{^the "([^"]+)" archive in the saved IUK should( not|) contain "([^"]+)"$}, func($c) {
    my $archive_name = $c->matches->[0];
    my $should_contain = $c->matches->[1] eq ' not' ? 0 : 1;
    my $expected_file = $c->matches->[2];

    $should_contain ?
        ok(archive_in_iuk_contains(
            $c->{stash}->{scenario}->{iuk_in},
            $archive_name,
            $expected_file,
        ))
        :
        ok(! archive_in_iuk_contains(
            $c->{stash}->{scenario}->{iuk_in},
            $archive_name,
            $expected_file,
        ))
};

Then qr{^the delete_files list should(| not) contain "([^"]+)"$}, func($c) {
    my $wanted            = $c->matches->[0] eq ' not' ? 0 : 1;
    my $expected_filename = $c->matches->[1];

    is(
        scalar(grep {
            $_ eq $expected_filename
        } @{$c->{stash}->{scenario}->{iuk_in}->delete_files}),
        $wanted
    );
};

Then qr{^the delete_files list should be empty$}, func($c) {
    is($c->{stash}->{scenario}->{iuk_in}->delete_files_count, 0);
};

Then qr{^the saved IUK should contain a SquashFS that contains file "([^"]+)"(?:| modified at ([0-9]+))$}, func($c) {
    my $expected_file  = $c->matches->[0];
    my $expected_mtime = $c->matches->[1];

    ok(squashfs_in_iuk_contains(
        $c->{stash}->{scenario}->{iuk_in},
        $c->{stash}->{scenario}->{squashfs_diff_name},
        $expected_file, $expected_mtime,
    ));
};

Then qr{^the "([^"]+)" archive in the saved IUK should contain a SquashFS diff$}, func($c) {
    my $archive_name = $c->matches->[0];

    ok(archive_in_iuk_contains(
        $c->{stash}->{scenario}->{iuk_in},
        $archive_name,
        file('live', $c->{stash}->{scenario}->{squashfs_diff_name}),
    ));
};

Then qr{^the saved IUK should contain a SquashFS that deletes file "([^"]+)"$}, func($c) {
    my $deleted_file = $c->matches->[0];

    ok(squashfs_in_iuk_deletes(
        $c->{stash}->{scenario}->{iuk_in},
        $c->{stash}->{scenario}->{squashfs_diff_name},
        $deleted_file
    ));
};

Then qr{^the "([^"]+)" file in the saved IUK is compressed with bzip2$}, func($c) {
    my $file = $c->matches->[0];
    my $tempdir = tempdir(CLEANUP => 1);
    my $orig_cwd = getcwd;
    chdir $tempdir;
    system(qw{tar xf}, $c->{stash}->{scenario}->{iuk_in}->file);
    ${^CHILD_ERROR_NATIVE} == 0 or croak("Failed to extract IUK");
    system('bunzip2', '--quiet', $file);
    my $ret = ${^CHILD_ERROR_NATIVE};
    chdir $orig_cwd;
    is($ret, 0);
};

Then qr{^the saved IUK should contain the new bootloader configuration$}, func($c) {
    my $tempdir = tempdir(CLEANUP => 1);
    my $orig_cwd = getcwd;
    chdir $tempdir;
    system(qw{tar xf}, $c->{stash}->{scenario}->{iuk_in}->file);
    ${^CHILD_ERROR_NATIVE} == 0 or croak("Failed to extract IUK");
    system(qw{bunzip2 --quiet boot.tar.bz2});
    ${^CHILD_ERROR_NATIVE} == 0 or croak("Failed to uncompress boot.tar.bz2");
    my $boot_tar = Archive::Tar->new('boot.tar');
    my $content = $boot_tar->get_content('syslinux/live.cfg');
    chdir $orig_cwd;
    is($content, 'new');
};

Then qr{^the "([^"]+)" archive should contain an upgraded syslinux configuration$}, func($c) {
    my $archive_name = $c->matches->[0];
    assert(defined $archive_name && length $archive_name);

    my @wanted_files   = qw{syslinux/syslinux.cfg EFI/BOOT/bootx64.efi utils/linux/syslinux};
    my @unwanted_files = qw{syslinux/isolinux.cfg};

    map {
        ok(archive_in_iuk_contains(
            $c->{stash}->{scenario}->{iuk_in}, $archive_name, $_,
        ), "archive '$archive_name' contains '$_'");
    } @wanted_files;

    map {
        ok(! archive_in_iuk_contains(
            $c->{stash}->{scenario}->{iuk_in}, $archive_name, $_,
        ), "archive '$archive_name' does not contain '$_'");
    } @unwanted_files;

    ok(file_content_in_archive_in_iuk_like(
            $c->{stash}->{scenario}->{iuk_in},
            $archive_name,
            "syslinux/syslinux.cfg",
            qr{/syslinux/}
    ), "syslinux/syslinux.cfg contains /syslinux/");

    ok(file_content_in_archive_in_iuk_unlike(
            $c->{stash}->{scenario}->{iuk_in},
            $archive_name,
            "syslinux/syslinux.cfg",
            qr{/isolinux/}
    ), "syslinux/syslinux.cfg does not contain /isolinux/");

};
