#!perl

use strict;
use warnings FATAL => 'all';

use lib qw{lib t/lib};

use Carp;
use Carp::Assert;
use Carp::Assert::More;
use Cwd;
use Data::Dumper;
use English qw{-no_match_vars};
use Fcntl ':mode';
use HTTP::Server::Simple;
use Test::More;
use Test::BDD::Cucumber::StepFile;
use Method::Signatures::Simple;

use Path::Class;
use File::Temp qw{tempdir tempfile};
use Tails::IUK::Utils qw{get_temp_dir};
use Test::SslCertificates qw{generate_ssl_privkey generate_self_signed_ssl_cert populate_ssl_template};
use Test::Util;
use Test::WebServer::RedirectToHTTPS;
use Test::WebServer::Static;
use Test::WebServer::Static::SSL;

my $bindir = file(__FILE__)->dir->parent->parent->parent->subdir('bin')->absolute;

$ENV{HARNESS_ACTIVE} = 1;

Given qr{^a usable temporary directory$}, func($c) {
    my $tempdir = $c->{stash}->{scenario}->{tempdir} = get_temp_dir();
    ok(-d $tempdir);
    my $destdir = $c->{stash}->{scenario}->{destdir} = dir($tempdir, 'destdir');
    $destdir->mkpath;
    ok(-d $destdir);
};

Given qr{^(a|another) random port$}, func($c) {
    my $port = 40000 + int(rand(10000));
    my $key  = $c->matches->[0] eq 'a' ? 'port' : 'another_port';
    $c->{stash}->{scenario}->{server}->{$key} = $port;
    ok(defined $c->{stash}->{scenario}->{server}->{$key});
};

func prepare_webroot ($webroot, $filename, $present, $webdir, $spec_type, $spec_value) {
    my ($size, $content);
    if (defined $spec_type) {
        if ($spec_type eq 'content') {
            $content = $spec_value;
        }
        elsif ($spec_type eq 'size') {
            $size = $spec_value;
        }
    }

    $webroot->mkpath;
    ok(-d $webroot->stringify);

    if ($present) {
        my $file = file($webroot, $webdir, $filename);
        $file->dir->mkpath;
        -d $file->dir || croak("Directory ", $file->dir, " was not created.");
        assert(defined $content || defined $size);
        my $fh = $file->openw;
        if (defined $content) {
            print $fh $content;
        }
        else {
            print $fh "c" x $size;
        }
        close $fh;
        ok(-e $file->stringify);
        if (defined $content) {
            is($file->slurp, $content);
        }
        else {
            is(-s $file, $size);
        }
    }
}

Given qr{^a HTTP server that(| does not) serve[s]? "([^"]+)" in "([^"]+)"(?: with (content|size) "?([^"]*)"?)?}, func($c) {
    my $present    = $c->matches->[0] ? 0 : 1;
    my $filename   = $c->matches->[1];
    my $webdir     = $c->matches->[2];
    my $spec_type  = $c->matches->[3];
    my $spec_value = $c->matches->[4];

    my $webroot = dir($c->{stash}->{scenario}->{tempdir}, 'webroot');
    prepare_webroot($webroot, $filename, $present, $webdir, $spec_type, $spec_value);

    my $port  = $c->{stash}->{scenario}->{server}->{port};
    my $s     = Test::WebServer::Static->new({webroot => $webroot}, $port);
    is($s->port(), $port, "Constructor set port correctly");
    my $pid   = $c->{stash}->{scenario}->{server}->{http_pid}  = $s->background();
    like($pid, '/^-?\d+$/', 'PID is numeric');
};

Given qr{^a HTTP server that redirects to ([^ ]+) over HTTPS$}, func($c) {
    my $target_hostname = $c->matches->[0];

    my $port  = $c->{stash}->{scenario}->{server}->{port};

    my $target = sprintf(
        "%s:%s",
        $target_hostname,
        $c->{stash}->{scenario}->{server}->{another_port}
    );

    my $s = Test::WebServer::RedirectToHTTPS->new({ target => $target }, $port);
    is($s->port(), $port, "Constructor set port correctly");
    my $pid   = $c->{stash}->{scenario}->{server}->{http_pid}  = $s->background();
    like($pid, '/^-?\d+$/', 'PID is numeric');
};

Given qr{^a HTTPS server (?:|on ([^ ]+)) that(| does not) serve[s]? "([^"]+)" in "([^"]+)"(?: with (content|size) "?([^"]*)"?)?}, func($c) {
    my $listen     = $c->matches->[0];
    my $present    = $c->matches->[1] ? 0 : 1;
    my $filename   = $c->matches->[2];
    my $webdir     = $c->matches->[3];
    my $spec_type  = $c->matches->[4];
    my $spec_value = $c->matches->[5];
    $listen //= '127.0.0.1';

    my $webroot = dir($c->{stash}->{scenario}->{tempdir}, 'webroot');
    prepare_webroot($webroot, $filename, $present, $webdir, $spec_type, $spec_value);

    # We don't rely on SSL certificates for target files security,
    # so for more robustness, we need to ensure they don't get
    # in the way for dumb reasons; basically, we want to skip validation
    # in most cases => we try to test the worst case, that is a self-signed
    # certificate, whose hostname does not match the server's one.
    # It would be even better if it was expired, but apparently certtool
    # does not know how to do it, and mocking the system time for OpenSSL
    # does not look that easy.

    my $key      = file($c->{stash}->{scenario}->{tempdir}, "ssl_privkey") ->stringify;
    my $cert     = file($c->{stash}->{scenario}->{tempdir}, "ssl_cert")    ->stringify;
    my $template = file($c->{stash}->{scenario}->{tempdir}, "ssl_template")->stringify;

    populate_ssl_template({ outfile => $template });
    generate_ssl_privkey({ outfile => $key });
    generate_self_signed_ssl_cert({
        outfile => $cert, privkey => $key, template => $template,
    });

    my $port  = $c->{stash}->{scenario}->{server}->{another_port};
    my $s     = Test::WebServer::Static::SSL->new(
        { webroot => $webroot, cert => $cert, key => $key, },
        $port
    );
    is($s->port(), $port, "Constructor set port correctly");
    $s->host($listen);
    is($s->host(), $listen, "Constructor set host correctly");
    my $pid   = $c->{stash}->{scenario}->{server}->{https_pid}  = $s->background();
    like($pid, '/^-?\d+$/', 'PID is numeric');
};

When qr{^I download "([^"]+)" \(of expected size (\d+)\) from "([^"]+)", and check its hash is "([^"]+)"$}, func($c) {
    my $filename      = $c->matches->[0];
    my $expected_size = $c->matches->[1];
    my $webdir        = $c->matches->[2];
    my $expected_hash = $c->matches->[3];

    my $uri = sprintf("%s:%d/%s/%s",
                      "127.0.0.1",
                      $c->{stash}->{scenario}->{server}->{port},
                      $webdir,
                      $filename);
    $uri =~ s{//}{/}gxms;
    $uri = "http://$uri";

    my $output_filename
        = $c->{stash}->{scenario}->{output_filename}
        = file($c->{stash}->{scenario}->{destdir}, 'output.file');

    my $cmdline =
        file($bindir, "tails-iuk-get-target-file") .
        ' --uri "'         . $uri             . '"' .
        ' --hash-type "'   . 'sha256'         . '"' .
        ' --hash-value "'  . $expected_hash   . '"' .
        ' --output-file "' . $output_filename . '"' .
        ' --size "'        . $expected_size   . '"'
    ;
    $c->{stash}->{scenario}->{output} = `umask 077 && $cmdline 2>&1`;
    $c->{stash}->{scenario}->{exit_code} = ${^CHILD_ERROR_NATIVE};
};

Then qr{^it should succeed$}, func($c) {
    Test::Util::kill_httpd($c);

    ok(defined $c->{stash}->{scenario}->{exit_code})
        and
    is($c->{stash}->{scenario}->{exit_code}, 0);

    if (defined $c->{stash}->{scenario}->{exit_code}
            && $c->{stash}->{scenario}->{exit_code} != 0
            && exists $c->{stash}->{scenario}->{output}
            && defined $c->{stash}->{scenario}->{output}) {
        warn $c->{stash}->{scenario}->{output};
    }
};

Then qr{^it should fail$}, func($c) {
    Test::Util::kill_httpd($c);

    ok(defined $c->{stash}->{scenario}->{exit_code})
        and
    isnt($c->{stash}->{scenario}->{exit_code}, 0);
};

Then qr{^I should be told "([^"]+)"$}, func($c) {
    my $expected_err = $c->matches->[0];
    like($c->{stash}->{scenario}->{output}, qr{$expected_err});
};

Then qr{^I should(| not) see the downloaded file in the temporary directory$}, func($c) {
    my $wanted   = $c->matches->[0] ? 0 : 1;

    my $output_filename = $c->{stash}->{scenario}->{output_filename};
    $wanted ? ok(-e $output_filename) : ok(! -e $output_filename);
};

Then qr{^the SHA-256 of the downloaded file should be "([^"]+)"$}, func($c) {
    my $expected_hash = $c->matches->[0];

    my $output_filename = $c->{stash}->{scenario}->{output_filename};
    assert(-e $output_filename);

    my $hash_line = `sha256sum "$output_filename"`;
    assert_is(${^CHILD_ERROR_NATIVE}, 0);

    my @hash_words = split(/\s+/, $hash_line);
    is($expected_hash, $hash_words[0]);
};

Then qr{^the downloaded content should be not be much bigger than (\d+)$}, func($c) {
    my $expected_size = $c->matches->[0];

    my $not_much_bigger_base   = 4096;
    my $not_much_bigger_factor = 1.20;

    my ($downloaded_size) =
        ( $c->{stash}->{scenario}->{output} =~
              m{was downloaded but its size \((\d+)\) should be} );
    assert(defined($downloaded_size));
    ok($downloaded_size < $not_much_bigger_base + $expected_size * $not_much_bigger_factor );
};

Then qr{^the downloaded file should be world-readable$}, func($c) {
    my $output_filename = $c->{stash}->{scenario}->{output_filename};
    assert(-e $output_filename);

    my $mode       = (stat($output_filename))[2];
    my $other_read = $mode & S_IROTH;
    ok($other_read);
};
