=head1 NAME

Tails::IUK::UpgradeDescriptionFile - describe and manipulate a Tails upgrade-description file

=cut

package Tails::IUK::UpgradeDescriptionFile;

use Moose;
use MooseX::Method::Signatures;
use MooseX::Types::Moose qw( :all );
use MooseX::Types::Path::Class;
use MooseX::Has::Sugar::Saccharin;

use 5.10.1;
use namespace::autoclean;
use warnings FATAL => 'all';

use autodie qw(:all);
use Carp;
use Carp::Assert;
use Carp::Assert::More;
use Data::Dumper;
use Dpkg::Version qw{version_compare};
use English qw{-no_match_vars};
use List::MoreUtils qw{any};
use List::Util qw{sum};
use Path::Class;
use YAML::Any;


=head1 ATTRIBUTES

=cut

has "$_"    => required ro Str, predicate => "has_$_"
    for (qw{product_name product_version build_target channel});

has upgrades =>
    lazy_build ro ArrayRef,
    traits => [ 'Array' ],
    handles => {
        count_upgrades => 'count',
        all_upgrades   => 'elements',
        add_upgrade    => 'push',
        empty_upgrades => 'clear',
    };

has upgrade_paths =>
    lazy_build ro ArrayRef,
    traits => [ 'Array' ],
    handles => {
        count_upgrade_paths => 'count',
        all_upgrade_paths   => 'elements',
    };


=head1 CONSTRUCTORS AND BUILDERS

=cut

method _build_upgrades { return [] }

method _build_upgrade_paths {
    my @upgrade_paths;
    foreach my $upgrade ($self->all_upgrades) {
        exists $upgrade->{'upgrade-paths'} or $upgrade->{'upgrade-paths'} = [];
        foreach my $path (@{$upgrade->{'upgrade-paths'}}) {
            foreach my $key (qw{type target-files}) {
                assert(exists  $path->{$key});
                assert(defined $path->{$key});
            }
            $path->{'details-url'}  = $upgrade->{'details-url'};
            $path->{'upgrade-type'} = $upgrade->{'type'};
            $path->{'version'}      = $upgrade->{'version'};
            $path->{'total-size'}   = sum(map { $_->{size} } @{$path->{'target-files'}});
            push @upgrade_paths, $path;
        }
    }
    return \@upgrade_paths;
}

sub new_from_text {
    my $class = shift;
    my $text  = shift;

    my $data = YAML::Any::Load($text);

    my %args;
    foreach my $key (qw{product-name product-version channel build-target upgrades}) {
        next unless exists $data->{$key};
        my $attribute = $key; $attribute =~ s{-}{_}xmsg;
        $args{$attribute} = $data->{$key};
    }

    $class->new(%args);
}

sub new_from_file {
    my $class    = shift;
    my $filename = shift;

    my $content = file($filename)->slurp;
    assert_nonblank($content);

    $class->new_from_text($content);
}


=head1 METHODS

=cut

method contains_upgrade_path () {
    $self->count_upgrade_paths > 0;
}

method incremental_upgrade_paths () {
    grep { $_->{type} eq 'incremental' } $self->all_upgrade_paths;
}

method full_upgrade_paths () {
    grep { $_->{type} eq 'full' } $self->all_upgrade_paths;
}

method contains_incremental_upgrade_path () {
    $self->incremental_upgrade_paths > 0;
}

method contains_full_upgrade_path () {
    $self->full_upgrade_paths > 0;
}

method incremental_upgrade_path () {
    path_to_newest_version($self->incremental_upgrade_paths);
}

method full_upgrade_path () {
    path_to_newest_version($self->full_upgrade_paths);
}

method stringify () {
    my %data;

    foreach my $attribute (qw{product_name product_version channel build_target upgrades}) {
        my $predicate = "has_$attribute";
        next unless $self->$predicate;
        my $key = $attribute; $key =~ s{_}{-}xmsg;
        $data{$key} = $self->$attribute;
    }

    YAML::Any::Dump(\%data);
}


=head1 FUNCTIONS

=cut

sub path_to_newest_version {
    my @paths = @_;
    assert(@paths);

    my $current_best_path = { version => '-1' };

    foreach my $path (@paths) {
        $current_best_path = $path
            if version_compare($path->{version}, $current_best_path->{version}) == 1;
    }

    return $current_best_path;
}

no Moose;
1;
