=head1 NAME

Tails::IUK::UpgradeDescriptionFile::Download - download and verify an upgrade description file

=cut

package Tails::IUK::UpgradeDescriptionFile::Download;

use Moose;
use Moose::Util::TypeConstraints qw{class_type};
use MooseX::Method::Signatures;
use MooseX::Types::Moose qw( :all );
use MooseX::Types::Path::Class;
use MooseX::Has::Sugar::Saccharin;



with 'MooseX::Getopt::Dashes';

use 5.10.1;
use namespace::autoclean;
use warnings FATAL => 'all';

use autodie qw(:all);
use Carp;
use Carp::Assert;
use Carp::Assert::More;
use Data::Dumper;
use English qw{-no_match_vars};
use HTTP::Request;
use IO::Handle;
use Path::Class;
use Tails::RunningSystem;
use Tails::IUK::Utils;
use WWW::Curl::Easy;
use YAML::Any;


=head1 TYPES

=cut

class_type 'Tails::RunningSystem';


=head1 ATTRIBUTES

=cut

has 'max_download_size' => lazy_build ro Int;
has "$_" => lazy_build ro Str
    for (qw{override_baseurl override_build_target override_os_release_file
            trusted_gnupg_homedir});
has 'running_system' =>
    lazy_build ro 'Tails::RunningSystem',
    handles => [
        qw{upgrade_description_file_url upgrade_description_sig_url},
        qw{product_name product_version build_target channel}
    ];
has 'curl_opts' => lazy_build ro 'HashRef';


=head1 CONSTRUCTORS AND BUILDERS

=cut

method _build_max_download_size { 8 * 2**10 }

method _build_trusted_gnupg_homedir     {
    my $trusted_gnupg_homedir = '/usr/share/tails-iuk/trusted_gnupg_homedir';
    assert(-d $trusted_gnupg_homedir);
    return $trusted_gnupg_homedir;
}

method _build_running_system {
    my @args;
    for (qw{baseurl build_target os_release_file}) {
        my $attribute = "override_$_";
        my $predicate = "has_$attribute";
        if ($self->$predicate) {
            push @args, ($_ => $self->$attribute)
        }
    }
    Tails::RunningSystem->new(@args);
}

method _build_curl_opts {
    my @opts = (
        CURLOPT_NOPROGRESS,      1,
        # This does *not* prevent curl from downloading more data this in the end.
        CURLOPT_MAXFILESIZE,     $self->max_download_size,
        CURLOPT_SSLVERSION,      CURL_SSLVERSION_TLSv1,
        CURLOPT_SSL_CIPHER_LIST, 'ECDHE-RSA-AES256-GCM-SHA384:ECDHE-RSA-AES256-SHA384:ECDHE-RSA-AES128-GCM-SHA256:ECDHE-RSA-AES128-SHA256:ECDHE-RSA-AES256-SHA:!RC4:HIGH:!MD5:!aNULL:!EDH',
    );
    if ($ENV{HARNESS_ACTIVE} or $ENV{DISABLE_PROXY}) {
        push @opts, CURLOPT_PROXY,     '';
    }
    else {
        push @opts, CURLOPT_PROXYTYPE, CURLPROXY_SOCKS5_HOSTNAME;
        push @opts, CURLOPT_PROXY,     '127.0.0.1:9062';
    }
    if ($ENV{SSL_NO_VERIFY}) {
        push @opts, CURLOPT_SSL_VERIFYHOST, 0;
        push @opts, CURLOPT_SSL_VERIFYPEER, 0;
    }
    else {
        my $cafile = $ENV{HTTPS_CA_FILE};
        $cafile  //= '/usr/local/etc/ssl/certs/tails.boum.org-CA.pem';
        push @opts, CURLOPT_SSL_VERIFYHOST,  2;
        push @opts, CURLOPT_SSL_VERIFYPEER,  1;
        push @opts, CURLOPT_CAINFO,          $cafile;
        push @opts, CURLOPT_PROTOCOLS,       CURLPROTO_HTTPS;
        push @opts, CURLOPT_REDIR_PROTOCOLS, CURLPROTO_HTTPS;
    }
    my %opts = @opts;
    return \%opts;
}


=head1 METHODS

=cut

method fatal (@msg) {
    Tails::IUK::Utils::fatal(msg => \@msg);
}

=head2 get_url

Returns decoded content found at URL.
Throws an exception on detected failure.

=cut

method get_url ($url) {
    my $curl  = WWW::Curl::Easy->new;
    $curl->setopt(CURLOPT_URL, $url);
    while (my ($k, $v) = each(%{$self->curl_opts})) {
        $curl->setopt($k, $v);
    }

    my $response_body;
    $curl->setopt(CURLOPT_WRITEDATA, \$response_body);
    my $retcode = $curl->perform;

    my $response_code;
    if ($retcode == 0) {
        $response_code = $curl->getinfo(CURLINFO_HTTP_CODE);
    } else {
        croak(sprintf(
            "Could not download '%s', request failed (%s): %s\n",
            $url, $curl->strerror($retcode), $curl->errbuf));
    }

    $response_code == 200 or croak(sprintf(
        "Could not download '%s', request failed (%s): %s\n",
        $url, $curl->strerror($retcode), $curl->errbuf,
    ));

    assert(defined $response_body);
    length $response_body or croak(sprintf(
        "Downloaded empty file at '%s'\n", $url
    ));

    length $response_body <= $self->max_download_size or croak(sprintf(
        "Downloaded from '%s' but the downloaded content (%d) should be smaller than %d",
        $url, length($response_body), $self->max_download_size,
    ));

    return $response_body;
}

method matches_running_system ($description_str) {
    assert(defined $description_str);
    my $description = YAML::Any::Load($description_str);
    assert_hashref($description);
    foreach (qw{product_name product_version build_target channel}) {
        my $accessor = my $field = $_;
        $field =~ s{_}{-}gxms;
        exists  $description->{$field}             or return;
        defined $description->{$field}             or return;
        $description->{$field} eq $self->$accessor or return;
    }
    return 1;
}

method run () {
    my $description = $self->get_url($self->upgrade_description_file_url);
    my $signature   = $self->get_url($self->upgrade_description_sig_url );

    verify_signature($description, $signature, $self->trusted_gnupg_homedir)
        or croak("Invalid signature");
    $self->matches_running_system($description)
        or croak("Does not match running system");
    print $description;
    return;
}

no Moose;
1;
