package Test::Util;

use strict;
use warnings FATAL => 'all';
use 5.10.1;

use Carp;
use English qw{-no_match_vars};
use List::Util qw{first};
use Method::Signatures::Simple;
use Path::Class;
use Sys::Filesystem ();
use Tails::IUK::Utils qw{run_as_root stdout_as_root};
use Test::More;

func kill_httpd ($c) {
    foreach my $pid_key (qw{http_pid https_pid}) {
        my $pid = $c->{stash}->{scenario}->{server}->{$pid_key};
        is(kill(9, $pid), 1, "Killed PID $pid" ) if defined $pid;
    }
}

func upgrade_description_header ($product_name, $product_version, $build_target, $channel) {
    return <<EOF
product-name: $product_name
product-version: $product_version
channel: $channel
build-target: $build_target
EOF
;
}

# $mount may be a device or mountpoint
func mount_options ($mount) {
    my $fs = Sys::Filesystem->new();
    return split(/,/, $fs->options($mount));
}

# $mount may be a device or mountpoint
func has_mount_option ($mount, $option) {
    grep { $_ eq $option } mount_options($mount);
}

func remount_for_root_ro ($device, $mountpoint) {
    run_as_root('umount', '-l', $mountpoint);
    ${^CHILD_ERROR_NATIVE} == 0 or croak("Failed to umount '$mountpoint'.");

    run_as_root('mount', '-o', 'umask=0022', $device, $mountpoint);
    ${^CHILD_ERROR_NATIVE} == 0 or croak("Failed to mount '$mountpoint'.");

    run_as_root('mount', '-o', 'remount,ro', $mountpoint);
    ${^CHILD_ERROR_NATIVE} == 0 or croak("Failed to remount '$mountpoint' read-only.");
}

func remount_for_me_rw ($device, $mountpoint) {
    run_as_root('umount', '-l', $mountpoint);
    ${^CHILD_ERROR_NATIVE} == 0 or croak("Failed to umount '$mountpoint'.");

    my @gids = split(/ /, $GID);
    run_as_root('mount', '-o', "rw,uid=$UID,gid=$gids[0]", $device, $mountpoint);
    ${^CHILD_ERROR_NATIVE} == 0 or croak("Failed to mount '$mountpoint'.");
}

# $size is in MB
func prepare_live_device ($backing_file, $mountpoint, $size) {
    $size //= 64;

    system(qw{dd if=/dev/zero bs=1M}, "of=$backing_file", "count=$size");
    system(
        qw{/sbin/sgdisk --clear --new=1:0:0},
        qw{--typecode=1:C12A7328-F81F-11D2-BA4B-00A0C93EC93B --change-name=Tails},
        $backing_file,
    );
    run_as_root(qw{kpartx -a -s}, $backing_file);

    # output looks like "loop0p1 : 0 128991 /dev/loop0 2048"
    my $kpartx_output = stdout_as_root(qw{kpartx -l}, $backing_file);
    my ($system_partition_basename, $boot_device) = (
        $kpartx_output =~ m{
                               \A (loop [0-9]+ p [0-9]+)
                               \s+ [:] \s+ [0-9]+ \s+ [0-9]+ \s+
                               (/dev/loop [0-9]+)
                               \s+ \d+
                       }xms)
        or croak("Failed to parse kpartx output:\n$kpartx_output");
    my @dmsetup_ls_output = stdout_as_root(qw{dmsetup ls});
    my $dm_line = first {
        m{^ $system_partition_basename \s+ [(] 25[34] : \d+ [)]$}xms
    } @dmsetup_ls_output
        or croak("Failed to parse dmsetup ls output.");
    my ($system_partition_dm_id) = (
        $dm_line =~ m{^ $system_partition_basename \s+ [(] 25[34] : (\d+) [)]$}xms
    ) or croak("Failed to parse dmsetup ls line output:\n$dm_line");
    my $system_partition = "/dev/dm-$system_partition_dm_id";

    run_as_root(qw{mkdosfs -F 32}, $system_partition);

    dir($mountpoint)->mkpath;
    my @gids = split(/ /, $GID);
    run_as_root(
        "mount", "-o", "uid=$UID,gid=$gids[0],umask=0022", '-t', 'vfat',
        $system_partition, $mountpoint
    );

    dir($mountpoint, 'live')->mkpath;
    my $squashfs = file($mountpoint, 'live', 'filesystem.squashfs');
    $squashfs->touch;
    my $modules_file = file($mountpoint, 'live', 'Tails.module');
    my $modules_fh = $modules_file->openw;
    say $modules_fh 'filesystem.squashfs';
    close $modules_fh;
    my $syslinux_dir = $mountpoint->subdir('syslinux');
    $syslinux_dir->mkpath;

    remount_for_root_ro($system_partition, $mountpoint);

    ok(-e $squashfs);
    ok(-e $modules_file);
    ok(-d $syslinux_dir);
    ok(  has_mount_option($mountpoint, 'ro'));
    ok(! has_mount_option($mountpoint, 'rw'));

    return (
        boot_device      => $boot_device,
        system_partition => $system_partition
    );
}

1;
