use Test::Spec;

use Env qw{@PATH};
use FindBin;
unshift @PATH, "$FindBin::Bin/../../bin";

use 5.10.1;
use Archive::Tar;
use Cwd;
use Data::Dumper;
use File::Temp qw{tempdir tempfile};
use Method::Signatures::Simple;
use Path::Class;
use Tails::IUK::Read;
use Tails::IUK::Utils qw{make_archive_with_files_called make_iuk_with_files};
use Test::Fatal qw{dies_ok};

func create_file($size) {
    my ($included_file_fh, $included_file) = tempfile;
    print $included_file_fh "c" x $size;
    close $included_file_fh;
    return $included_file;
}

describe 'A read IUK object' => sub {
    describe 'whose constructor is built with no arguments' => sub {
        it 'dies' => sub {
            dies_ok {Tails::IUK::Read->new() };
        };
    };
    describe 'has a "new_from_file" method that' => sub {
        describe 'if called on an IUK file that contains no FORMAT file' => sub {
            my ($iuk_fh, $iuk_filename);
            before sub {
                ($iuk_fh, $iuk_filename) = tempfile();
                file($iuk_filename)->touch;
            };
            it 'should die' => sub {
                dies_ok { Tails::IUK::Read->new_from_file($iuk_filename) };
            };
        };
    };
    describe 'has a "contains_file" method that' => sub {
        describe 'if the object is built from an IUK file that contains a whatever.file file' => sub {
            my $iuk;
            before sub {
                my $tempdir = tempdir(CLEANUP => 1);
                my ($iuk_fh, $iuk_filename) = tempfile();
                my $tar = Archive::Tar->new;
                {
                    my $orig_cwd = getcwd;
                    chdir $tempdir;
                    my $fh = file('FORMAT')->openw;
                    print $fh "1";
                    close $fh;
                    $tar->add_files('FORMAT');
                    file('whatever.file')->touch;
                    $tar->add_files('whatever.file');
                    chdir $orig_cwd;
                }
                $tar->write($iuk_filename);
                $iuk = Tails::IUK::Read->new_from_file($iuk_filename);
            };
            it 'should return true when called on "whatever.file"' => sub {
                ok($iuk->contains_file("whatever.file"));
            };
        };
        describe 'if the object is built from a valid IUK file that contains no whatever.file file' => sub {
            my $iuk;
            my ($iuk_fh, $iuk_filename);
            before sub {
                my $tempdir = tempdir(CLEANUP => 1);
                ($iuk_fh, $iuk_filename) = tempfile();
                my $tar = Archive::Tar->new;
                {
                    my $orig_cwd = getcwd;
                    chdir $tempdir;
                    my $fh = file('FORMAT')->openw;
                    print $fh "1";
                    close $fh;
                    $tar->add_files('FORMAT');
                    chdir $orig_cwd;
                }
                $tar->write($iuk_filename);
                $iuk = Tails::IUK::Read->new_from_file($iuk_filename);
            };
            it 'should return false when called on "whatever.file"' => sub {
                ok(! $iuk->contains_file("whatever.file"));
            };
        };
    };
    describe 'has a "delete_files" method that' => sub {
        describe 'if the object is built from a valid IUK whose delete_files is empty' => sub {
            my $iuk;
            before sub {
                my $tempdir = tempdir(CLEANUP => 1);
                my ($iuk_fh, $iuk_filename) = tempfile();
                my $tar = Archive::Tar->new;
                {
                    my $orig_cwd = getcwd;
                    chdir $tempdir;
                    my $fh;

                    $fh = file('FORMAT')->openw;
                    print $fh "1";
                    close $fh;
                    $tar->add_files('FORMAT');

                    $fh = file('control.yml')->openw;
                    print $fh "delete_files:\n";
                    close $fh;
                    $tar->add_files('control.yml');

                    chdir $orig_cwd;
                }
                $tar->write($iuk_filename);
                $iuk = Tails::IUK::Read->new_from_file($iuk_filename);
            };
            it 'should return an empty list' => sub {
                is(@{$iuk->delete_files}, 0);
            };
        };
        describe 'if the object is built from a valid IUK whose delete_files contains whatever.file' => sub {
            my $iuk;
            before sub {
                my $tempdir = tempdir(CLEANUP => 1);
                my ($iuk_fh, $iuk_filename) = tempfile();
                my $tar = Archive::Tar->new;
                {
                    my $orig_cwd = getcwd;
                    chdir $tempdir;
                    my $fh;

                    $fh = file('FORMAT')->openw;
                    print $fh "1";
                    close $fh;
                    $tar->add_files('FORMAT');

                    $fh = file('control.yml')->openw;
                    print $fh "delete_files:\n  - file1\n  - file2\n  - whatever.file\n  - file4\n";
                    close $fh;
                    $tar->add_files('control.yml');

                    chdir $orig_cwd;
                }
                $tar->write($iuk_filename);
                $iuk = Tails::IUK::Read->new_from_file($iuk_filename);
            };
            it 'should return a list that contains whatever.file' => sub {
                is(scalar(grep {$_ eq 'whatever.file'} @{$iuk->delete_files}), 1);
            };
        };
    };
    describe 'has a space_needed method that' => sub {
        describe 'if called on an IUK that contains no archive' => sub {
            my $iuk;
            before sub {
                my $tempdir = tempdir(CLEANUP => 1);
                my ($iuk_fh, $iuk_filename) = tempfile;
                close $iuk_fh;
                make_iuk_with_files($iuk_filename, $tempdir);
                $iuk = Tails::IUK::Read->new_from_file($iuk_filename);
            };
            it 'should return 0' => sub {
                is($iuk->space_needed, 0);
            };
        };
        describe 'if called on an IUK that contains one archive that contains two 1MB files' => sub {
            my $iuk;
            before sub {
                my $tempdir = tempdir(CLEANUP => 1);
                my ($iuk_fh, $iuk_filename) = tempfile; close $iuk_fh;
                my ($archive_fh, $archive_filename) = tempfile; close $archive_fh;
                rename $archive_filename, "$archive_filename.tar";
                $archive_filename = "$archive_filename.tar";
                make_archive_with_files_called(file($archive_filename), size => 1,
                                               filenames => ['whatever1.file', 'whatever2.file']);
                make_iuk_with_files($iuk_filename, $tempdir, $archive_filename);
                $iuk = Tails::IUK::Read->new_from_file($iuk_filename);
            };
            it 'should return 2 * 2**10' => sub {
                is($iuk->space_needed, 2 * 2**20);
            };
        };
    };
};

runtests unless caller;
