use Test::Spec;

use 5.10.1;
use Archive::Tar;
use Data::Dumper;
use File::Temp qw{tempdir tempfile};
use Method::Signatures::Simple;
use Path::Class;
use Tails::IUK::Utils qw{unpacked_size};
use Test::Fatal qw{dies_ok};

func create_file($size) {
    my ($included_file_fh, $included_file) = tempfile;
    print $included_file_fh "c" x $size;
    close $included_file_fh;
    return $included_file;
}

describe 'The unpacked_size function' => sub {
    describe 'when run on a non-existing archive' => sub {
        it 'throws an exception' => sub {
            dies_ok { unpacked_size('non-existing.tar') };
        };
    };
    describe 'when run on an uncompressed archive that contains an empty file' => sub {
        my ($archive_fh, $archive_file) = tempfile;
        before sub {
            my $archive = Archive::Tar->new();
            my $included_file = create_file(0);
            $archive->add_files($included_file);
            $archive->write($archive_file);
        };
        it 'returns 0' => sub {
            is(unpacked_size($archive_file), 0);
        };
    };
    describe 'when run on a compressed archive that contains an empty file' => sub {
        my ($archive_fh, $archive_file) = tempfile;
        before sub {
            my $archive = Archive::Tar->new();
            my $included_file = create_file(0);
            $archive->add_files($included_file);
            $archive->write($archive_file, COMPRESS_BZIP);
        };
        it 'returns 0' => sub {
            is(unpacked_size($archive_file), 0);
        };
    };
    describe 'when run on a compressed archive that contains a 1kB file' => sub {
        my ($archive_fh, $archive_file) = tempfile;
        before sub {
            my $archive = Archive::Tar->new();
            my $included_file = create_file(1024);
            $archive->add_files($included_file);
            $archive->write($archive_file, COMPRESS_BZIP);
        };
        it 'returns 1024' => sub {
            is(unpacked_size($archive_file), 1024);
        };
    };
    describe 'when run on a compressed archive that contains two 1kB files' => sub {
        my ($archive_fh, $archive_file) = tempfile;
        before sub {
            my $archive = Archive::Tar->new();
            my $included_file1 = create_file(1024);
            my $included_file2 = create_file(1024);
            $archive->add_files($included_file1, $included_file2);
            $archive->write($archive_file, COMPRESS_BZIP);
        };
        it 'returns 2048' => sub {
            is(unpacked_size($archive_file), 2048);
        };
    };
};

runtests unless caller;
